'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var core = require('@babel/core');
var util = require('util');
var fs = require('fs');
var path = require('path');
var resolve = require('resolve');

function assert(expression, message) {
  if (!expression) {
    for (var _len = arguments.length, args = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
      args[_key - 2] = arguments[_key];
    }

    throw new Error(util.format("babel-plugin-direct-import: ".concat(message), ...args));
  }
}
function assertNotNull(value, message) {
  for (var _len2 = arguments.length, args = new Array(_len2 > 2 ? _len2 - 2 : 0), _key2 = 2; _key2 < _len2; _key2++) {
    args[_key2 - 2] = arguments[_key2];
  }

  assert(value != null, message, ...args);
}

function resolveModule(fileName) {
  var basedir = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : process.cwd();

  try {
    return resolve.sync(fileName, {
      basedir,

      packageFilter(pkg) {
        pkg.main = pkg.module || pkg.esnext || pkg['jsnext:main'] || pkg.main;
        return pkg;
      }

    });
  } catch (error) {
    if (error.code === 'MODULE_NOT_FOUND') {
      return null;
    }

    throw error;
  }
}

function toRelativeSource(filePath) {
  var pathChunks = filePath.split(path.sep);
  var relativeChunks = pathChunks.slice(pathChunks.lastIndexOf('node_modules') + 1);
  return relativeChunks.join('/');
}

function resolveIndexPath(name, indexFile) {
  if (!indexFile) {
    var _indexFilePath = resolveModule(name);

    assertNotNull(_indexFilePath, "failed to find 'indexFile' of '%s'.", name);
    return _indexFilePath;
  }

  var pkgJSON = resolveModule("".concat(name, "/package.json"));
  assertNotNull(pkgJSON, "failed to resolve package.json for '%s'", name);
  var pkgDir = path.dirname(pkgJSON);
  var indexFilePath = resolveModule(path.join(pkgDir, indexFile));
  assertNotNull(indexFilePath, "failed to resolve 'indexFile' specified for '%s'.", name);
  return indexFilePath;
}

function fulfillExports(exports, filePath) {
  var content = fs.readFileSync(filePath, 'utf-8');
  var ast = core.parse(content, {
    filename: filePath,
    sourceType: 'module'
  });
  var program = !ast ? null : ast.type === 'File' ? ast.program : ast;
  assertNotNull(program, "failed to parse '%s'.", filePath);
  var fileDir = path.dirname(filePath);
  var imports = new Map();

  for (var node of program.body) {
    if (core.types.isImportDeclaration(node)) {
      var sourcePath = resolveModule(node.source.value, fileDir);

      if (!sourcePath) {
        continue;
      }

      for (var specifier of node.specifiers) {
        imports.set(specifier.local.name, {
          source: toRelativeSource(sourcePath),
          imported: core.types.isImportNamespaceSpecifier(specifier) ? '*' : core.types.isImportDefaultSpecifier(specifier) ? 'default' : specifier.imported.name
        });
      }
    } else if (core.types.isExportNamedDeclaration(node)) {
      var nodeDeclaration = node.declaration;

      if (core.types.isVariableDeclaration(nodeDeclaration)) {
        for (var declaration of nodeDeclaration.declarations) {
          if (core.types.isIdentifier(declaration.id) && core.types.isIdentifier(declaration.init)) {
            var external = declaration.id.name;
            var internal = declaration.init.name;
            var imported = imports.get(internal);

            if (imported) {
              exports.set(external, {
                external,
                source: imported.source,
                internal: imported.imported
              });
            }
          }
        }
      }

      var _sourcePath = !node.source ? null : resolveModule(node.source.value, fileDir);

      for (var _specifier of node.specifiers) {
        if (_specifier.type === 'ExportSpecifier') {
          var _internal = _specifier.local.name;
          var _external = _specifier.exported.name;

          if (_sourcePath) {
            exports.set(_external, {
              internal: _internal,
              external: _external,
              source: toRelativeSource(_sourcePath)
            });
          } else {
            var _imported = imports.get(_internal);

            if (_imported) {
              exports.set(_external, {
                external: _external,
                source: _imported.source,
                internal: _imported.imported
              });
            }
          }
        }
      }
    } else if (core.types.isExportAllDeclaration(node)) {
      var _sourcePath2 = resolveModule(node.source.value, fileDir);

      assertNotNull(_sourcePath2, 'failed to resolve %s', _sourcePath2);
      fulfillExports(exports, _sourcePath2);
    }
  }
}

function getConfigExports(config) {
  var indexPath = resolveIndexPath(config.name, config.indexFile);
  var exports = new Map();
  fulfillExports(exports, indexPath);
  return exports;
}

function _objectWithoutProperties(source, excluded) {
  if (source == null) return {};

  var target = _objectWithoutPropertiesLoose(source, excluded);

  var key, i;

  if (Object.getOwnPropertySymbols) {
    var sourceSymbolKeys = Object.getOwnPropertySymbols(source);

    for (i = 0; i < sourceSymbolKeys.length; i++) {
      key = sourceSymbolKeys[i];
      if (excluded.indexOf(key) >= 0) continue;
      if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue;
      target[key] = source[key];
    }
  }

  return target;
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}
function prepareConfig(args) {
  var configs = Array.isArray(args) ? args : [args];
  return configs.map(options => {
    if (typeof options === 'string') {
      return {
        name: options
      };
    }

    var {
      name,
      indexFile
    } = options,
        unknownOptions = _objectWithoutProperties(options, ["name", "indexFile"]);

    assert(typeof name === 'string', '{ name } expected to be a string');
    assert(!!name, '{ name } is empty');

    if (indexFile != null) {
      assert(typeof indexFile === 'string', '{ indexFile } expected to be a string');
      assert(!!indexFile, '{ indexFile } is empty');
    }

    if (indexFile != null && indexFile !== name) {
      var [nameID] = name.split('/');
      var [indexFileID] = indexFile.split('/');
      assert(nameID === indexFileID, "index file '%s' must belong to '%s' package", indexFile, name);
    }

    var unknownOptionsKeys = Object.keys(unknownOptions);
    assert(unknownOptionsKeys.length === 0, 'contains unknown keys { %s }', unknownOptionsKeys.join(', '));
    return options;
  });
}

var configsCache = new Map();
var exportsCache = new Map();

function getConfigs(modules) {
  if (configsCache.size === 0) {
    for (var config of prepareConfig(modules)) {
      configsCache.set(config.name, config);
    }
  }

  return configsCache;
}

function getExports(name, modules) {
  var pkgExports = exportsCache.get(name);

  if (!pkgExports) {
    var config = getConfigs(modules).get(name);

    if (!config) {
      pkgExports = new Map();
    } else {
      pkgExports = getConfigExports(config);
    }

    exportsCache.set(name, pkgExports);
  }

  return pkgExports;
}

function plugin() {
  return {
    visitor: {
      ImportDeclaration(declaration, _ref) {
        var {
          opts
        } = _ref;
        var {
          source,
          specifiers,
          importKind
        } = declaration.node;

        if (!opts || !opts.modules || !Array.isArray(opts.modules) || opts.modules.length === 0) {
          return;
        }

        if (importKind === 'type') {
          return;
        }

        if (specifiers.length === 0) {
          return;
        }

        var exports = getExports(source.value, opts.modules);

        if (exports.size === 0) {
          return;
        }

        var _loop = function _loop(specifier) {
          if (core.types.isImportNamespaceSpecifier(specifier)) {
            // eslint-disable-next-line no-console
            console.warn(util.format('babel-plugin-direct-import: Can not optimize `import * as %s from "%s"`.'.concat('\n', 'See plugin limitations https://git.io/vFDOO for more details.'), specifier.local.name, source.value));
            return "continue";
          }

          var moduleName = core.types.isImportDefaultSpecifier(specifier) ? 'default' : specifier.imported.name;
          var moduleExports = exports.get(moduleName);

          if (moduleExports) {
            declaration.node.specifiers = declaration.node.specifiers.filter(x => x !== specifier);
            declaration.insertBefore(core.types.importDeclaration([moduleExports.internal === '*' ? core.types.importNamespaceSpecifier(core.types.identifier(specifier.local.name)) : moduleExports.internal === 'default' ? core.types.importDefaultSpecifier(core.types.identifier(specifier.local.name)) : core.types.importSpecifier(core.types.identifier(specifier.local.name), core.types.identifier(moduleExports.external))], core.types.stringLiteral(moduleExports.source)));
          }
        };

        for (var specifier of specifiers) {
          var _ret = _loop(specifier);

          if (_ret === "continue") continue;
        }

        if (declaration.node.specifiers.length === 0) {
          declaration.remove();
        }
      }

    }
  };
}

exports.default = plugin;
//# sourceMappingURL=index.js.map
